# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import tempfile
import datetime
import shlex
import subprocess
import os

from typing import Tuple, Dict, Any, List, Union, Optional

from tos_execute import config
from tos_execute import error_log
from tos_execute import models_internal as intmodels
from tos_execute import models_activeworkflow as awmodels


def prepare_job_execution(
    job_run: intmodels.JobRun,
) -> Tuple[intmodels.JobRun, List[str]]:
    """ Prepare the execution of a TOS job based on the JobRun and the context

    This functional method will prepare all necessary steps prior to a job execution and return 
    updated versions of a JobRun, the context dictionary as well as the run-command list.

    Parameters
    ----------
    job_run: intmodels.JobRun
        The JobRun instance that is stored as queued in the database.
    
    Returns
    -------
    intmodels.JobRun
        The JobRun instance with possible modifications according to the pre-process workflow.
    List[str]
        The job start shell command as a List
    """

    # Check if TOS output is set, create file path for it if necessary
    if job_run.tos_response_path is not None:
        talend_output_path = os.path.join(
            tempfile.gettempdir(),
            f"{job_run.job.name.lower()}_{job_run.job.version.lower()}_{job_run.id}.json",
        )
        job_run.context[job_run.tos_response_path] = talend_output_path
        job_run.tos_response_path = talend_output_path

    context_params = [
        f"--context_param={shlex.quote(param)}={shlex.quote(str(val))}"
        for param, val in job_run.context.items()
    ]

    os.chmod(job_run.job.path, 0o755)  # make the exec sript executable
    cmd = [f"{job_run.job.path}"] + context_params

    return (job_run, cmd)


def post_process_job_run(
    job_run: intmodels.JobRun, process: subprocess.CompletedProcess,
) -> intmodels.JobRun:
    """ Conduct all necessary steps to post-process a run job.

    This functional method takes all steps necessary to post process a run job based
    on the JobRun and the CompletedProcess instance. It will set some stats
    as well as the possible response message.

    Paramters
    ---------
    job_run: intmodels.JobRun
        The JobRun instance in started state.
    process: subprocess.CompletedProcess
        The CompletedProcess instance in which the TOS job has run.
    
    Returns
    -------
    intmodels.JobRun
        The JobRun instance with possible modifications according to the post-process workflow.
    """

    # set stats
    job_run.completed_at = datetime.datetime.now()
    job_run.state = intmodels.JobState.completed

    # set response
    response = awmodels.ResponseCheck()
    response.result.messages.append(
        awmodels.MessageOutput(
            job_run_id=job_run.id,
            job_id=job_run.job.id,
            job_name=job_run.job.name,
            job_version=job_run.job.version,
            job_parameters=job_run.job_parameters,
            start_time=job_run.started_at.isoformat(),
            end_time=job_run.completed_at.isoformat(),
        )
    )

    # set stdout / stderr
    if len(process.stdout) >= 1:
        response.result.logs = [process.stdout]
    if len(process.stderr) >= 1:
        # TODO: ugly ugly ugly!!
        cleaned = remove_slf4j_errors(process.stderr)
        if cleaned:
            response.result.errors = [cleaned]

    # set / calculate the job run stats
    response.result.messages[0].completed = True
    response.result.messages[0].returncode = process.returncode
    response.result.messages[0].runtime = (
        job_run.completed_at - job_run.started_at
    ).total_seconds()

    # set / retrieve the job run response
    if job_run.tos_response_path is not None:
        response.result.messages[0].response = retrieve_output_file(
            job_run.tos_response_path
        )
    else:
        response.result.messages[0].response = {}
    # hacking dirty: write every date as a JSON string and re-parse it as a dictionary for response.
    job_run.response = json.loads(response.json(exclude_none=True))

    return job_run


def retrieve_output_file(path: str) -> Union[Dict[Any, Any], List[Any]]:
    """ Retrieve a TOS job output from a path.

    Retrieves a reponse that has been written by the TOS job to path. Return
    this file as a pythonized JSON type (Dict or List).

    Parameters
    ----------
    path: str
        The full path to the file that has been written by a TOS job.
    
    Results
    -------
    Union[Dict[Any, Any], List[Any]]
        A pythonized JSON, read from the message written be a TOS job. Should
        be either Dict[Any, Any] (JSON Object) or a List[Any] (JSON Array)
    """
    response = {}
    if path != "" and os.path.exists(path) and os.path.isfile(path):
        try:
            with open(path, "r") as response_file:
                response = json.load(response_file)
        except:
            # Cannot read? no response then
            error_log.error(f"Could not read the output file: {path}")
            response = {}
    else:
        error_log.warning(f"Expected a file at path {path}, but it is invalid.")
    return response


def remove_slf4j_errors(stderr: Optional[str]) -> Optional[str]:
    """ Dirty hack method: remove all SLF4J errors from a String.

    Each line in the String will be checked for "SLF4J". If a line
    starts with this part, remove this line from the stderr input
    str. Return all other lines.

    Parameters
    ----------
    stderr: Optional[str]
        Any string. Should be a multiline string that is the stderr output
        from a CompletedProcess.
    
    Results
    -------
    Optional[str]
        The input string where every line starting with "SLF4J:" has been removed.
    """
    if not stderr:
        None
    out_str = ""
    for line in stderr.splitlines(keepends=True):
        if line.startswith("SLF4J:"):
            pass
        else:
            out_str += line
    return out_str


def is_authorized(
    params: Union[awmodels.ParamsRegister, awmodels.ParamsReceive, awmodels.ParamsCheck]
) -> bool:
    """ Check if a params section is authorized to use this TOS agent instance.

    This method will check 1) if credentials are set in the configuration of the agent and 2)
    whether the supplied params are authorized to use the agent. If not credentials are set
    (which is the default), the method will return True. In any other case, the `api_key`
    option has to contain the credential name in Liquid Templating notation which refers to
    the credentials parameter.

    Parameters
    ----------
    params: Union[awmodels.ParamsRegister, awmodels.ParamsReceive, awmodels.ParamsCheck]
        Any params dictionary posted by the requester JSON
    
    Results
    -------
    bool
        True if parameters hold valid credentials, authorizing any request. False otherwise
    """

    if len(config.BasicSettings().application_api_keys) == 0:
        # If there are no credentials set, no credentials are needed.
        return True

    set_api_key = params.options.api_key
    if set_api_key.startswith("{%") and set_api_key.endswith("%}"):
        set_api_key = set_api_key[2:-2].strip()
        if set_api_key.startswith("credential"):
            search_credential_name = set_api_key[len("credential") :].strip()
        else:
            # If you end up here: an api key like `{% something %}` has been supplied. Either use
            # the `credential` key word or use the plain API key, but do not mix.
            return False
    else:
        search_credential_name = set_api_key.strip()

    for credential in params.credentials:
        if credential.name == search_credential_name:
            if credential.value in config.BasicSettings().application_api_keys:
                return True

    return False
